; -------------------------------------
;   Z80 Return-oriented Virtual Machine 
;   (c) 2016 - 2017 kio@little-bat.de
; -------------------------------------


; Z80 Registers:
;
; AF  = 
; BC  = VM vstack (top down)
; DE  = TOP (1st value on vstack)
; HL  = AUX (2nd value on stack or high word of int32)
; IX  = 
; IY  = reserved
; AF' = 
; BC' = VM rstack (bottom up)
; DE' = 
; HL' = 
;
; values on stacks are always 2 bytes.
; stacks are aligned to even addresses.
; the high byte of local signed byte variables is invalid!

#assert VM_BC_IS_VSTACK = 1 	; variables stack pointer
#assert VM_HAS_RSTACK	= 1	; the VM uses a separate return stack
#assert VM_BC2_IS_RSTACK= 1	; return stack pointer
#assert VM_HAS_HSTACK	= 0	; the VM uses a separate stack for objects allocated on the heap
#assert VM_IY_IS_HSTACK	= 0	; variables stack pointer for heap objects
#assert VM_DE_IS_TOP	= 1	;
#assert VM_HL_IS_AUX	= 1	;

#assert VM_ALWAYS_EI	= 0	; interrupts may overwrite program code: must be restored from copy
#assert VM_ALWAYS_DI	= 1	; interrupts must not overwrite program code
#assert VM_POLL_IRPTS	= 0	; bra, call and ret poll interrupts by ei + di
#assert VM_POLL_IRPTS_K1BUS = 1	; bra, call and ret poll interrupts by testing K1-BUS status



#include "core-macros.s"
#local


#data data
sp_safe::	ds	2	; sp safe while sp = vsp in complex opcode implemenations
scratch::	ds	2*6	; scratch in VM opcodes, irpt-stack
irpt_stack::	ds	0	; top of interrupt stack


#code VM_CORE

; -----------------------------------------
; Utilities:
; -----------------------------------------

EXHLDE::	ex  hl,de	;4
NOP::		ret		;10 	14cc

POPHLDE::
POP32::		pop_de		;34
POPHL::		pop_hl		;34
		ret		;10	44cc / 78cc

; helper: fetch arguments for assignment operator:

LDIXDE_LDHLxDE_POPDE::
		ld	ix,de	;16
		ld	a,(de++);13
		ld	l,a	;4
		ld	a,(de)	;7
		ld	h,a	;4
		;pop_de		;34
		;ret		;10	= 88cc

POPDE::		pop_de		;34
		ret		;10	44cc


PUSHHLDE::
PUSH32::	push_hl		;34
PUSHDE::	push_de		;34
		ret		;10 	44cc / 78cc

PUSHHL::	push_hl		;34
		ret		;10 	44cc


; ________________________________
; load VM SP into IX
;
; use case:
; Z80 code needs simplified access to bytes somewhere in the vstack
;
; ld_ix_vsp ( -- )		
; ld_iy_vsp ( -- )		

.macro ld_ix_vsp		; 4 bytes + 16cc	uses ill. instr.
	ld	ix,bc
.endm	

.macro ld_iy_vsp		; 4 bytes + 16cc	uses ill. instr.
	ld	iy,bc
.endm	

		

; -----------------------------------------
; Code Flow:
; -----------------------------------------

; ________________________________
; switch from VM to native Z80 and back to VM:
;
; use case:
; execute some 	Z80 machine code in a VM-coded procedure.
; simple case: 	No stack: push, pop, call etc. not allowed!
; 		AF, HL, AF', DE', HL', IX and IY can be used.
; 		BC, DE, BC' and SP must be handled properly.
;
;	...	; virtual code
;	...
;	dw  NATIVE
;	...
;	...	; z80 code 
;	...
;	virtual
;	...
;	...	; virtual code
;
; NATIVE ( -- )		VM opcode	25cc
; virtual ( -- )	Z80 macro	20cc

NATIVE::
	ld	hl,0
	add	hl,sp		; HL = VM PC
	jp	hl		; PC = VM PC

.macro virtual
	ld	sp,$+4		; set VM PC
	ret			; next
.endm


; ________________________________
; switch from VM to native Z80 and back to VM:
;
; use case:
; execute some 	Z80 machine code in a VM-coded procedure.
; complex case:	With stack: push, pop, call etc. allowed!
; 		AF, BC, HL, AF', DE' and HL' can be used.
; 		DE and BC' must be handled properly.
;
;	...	; virtual code
;	...
;	dw  NATIVE_wSP
;	...
;	...	; z80 code 
;	...
;	virtual_wsp
;	...
;	...	; virtual code
;
; NATIVE_wstack ( de -- de )	VM opcode	49cc
; virtual_wstack ( de -- de )	Z80 macro	70cc

NATIVE_wstack::
	ld	ix,0		;14
	add	ix,sp		;15	IX = VM PC
	ld	hl,bc		;8	
	ld	sp,hl		;4	SP = VM vstack
	jp	ix		;8	= 49cc

.macro virtual_wstack
	ld	hl,0		;10
	add	hl,sp		;11
	ld	bc,hl		;8	VM SP = SP		
	ld	sp,$+4		;10	VM PC = PC
	ret			;10	= 49cc
.endm


; ________________________________
; load VM SP into SP and back into VM SP
; since the SP is the VM PC the SP must be temporarily saved somewhere
; 
; =>	stack becomes usable for call/ret/push/pop
;	arguments can be popped with "pop rr"
; 	register BC can be used
;
; .macro ld_sp_vsp_via_hl  ( ade -- ade )	 7 bytes: 32cc
; .macro ld_sp_vsp_via_ix  ( ahlde -- ahlde )	10 bytes: 44cc
; .macro ld_sp_vsp_via_ram ( ahlde -- ahlde )	12 bytes: 60cc
;
; .macro ld_vsp_sp_via_hl  ( ade -- ade )	10 bytes: 49cc
; .macro ld_vsp_sp_via_ix  ( ahlde -- ahlde )	14 bytes: 65cc  (nimp)
; .macro ld_vsp_sp_via_ram ( ahlde -- ahlde )	12 bytes: 60cc
;
; implementations of the restore macros:
; =>	can be jumped to
;	jump to next VM opcode == resume VM program 
;
; ld_vsp_sp_next      ( -- de )
; ld_vsp_sp_next_hlde ( -- hlde )

.macro ld_sp_vsp_via_hl		; 7 bytes + 32cc
	ld	(sp_safe),sp	;20
	ld	hl,bc		;8
	ld	sp,hl		;4
.endm

.macro ld_sp_vsp_via_ix		; 10 bytes + 44cc
	ld	(sp_safe),sp	;20
	ld	ix,bc		;16
	ld	sp,ix		;8
.endm

.macro ld_sp_vsp_via_ram	; 12 bytes + 60cc
	ld	(sp_safe),sp	;20
	ld	(scratch),bc	;20
	ld	sp,(scratch)	;20
.endm

.macro ld_vsp_sp_via_hl		; 10 bytes + 49cc
	ld	hl,0		;10
	add	hl,sp		;11
	ld	bc,hl		;8
	ld	sp,(sp_safe)	;20
.endm

;macro ld_vsp_sp_via_ix		; 14 bytes + 65cc	uses ill. instr.
;	ld	ix,0 		;14
;	add	ix,sp		;15
;	ld	bc,ix		;16
;	ld	sp,(sp_safe)	;20
;endm

.macro ld_vsp_sp_via_ram	; 12 bytes + 60cc
	ld	(scratch),sp	;20
	ld	bc,(scratch)	;20
	ld	sp,(sp_safe)	;20
.endm

;ld_vsp_sp_next::		; pres. de
;	ld_vsp_sp_via_hl	;49
;	ret			;10
;
;ld_vsp_sp_next_hlde::		; pres hlde
;	ld_vsp_sp_via_ram	;60
;	ret			;10


; ________________________________________________________
; Enter VM coded procedure:
; procedures are entered in native Z80 code!
; .macro p_enter pushes the VM PC on the VM rstack and resumes in VM mode
; the procedure must return with RETURN (or variant)
;
; foo::	p_enter			; 95cc
;	...
;	dw  RETURN		; 56cc
;
; .macro p_enter    ( de -- de )	; 6 bytes: 95cc
; .macro p_enter_hlde ( hlde -- hlde )	; 7 bytes: 103cc  NIMP (saving 4 bytes is worth 20cc)
; .macro p_enter_hlde ( hlde -- hlde )	; 3 bytes: 123cc  _must_ be first instr of proc!

#if VM_BYTE_OFFSETS ; short VM program wanted?
.macro p_enter
	jp	p_enter_hlde	;10
.endm
#else ; !VM_BYTE_OFFSETS
.macro p_enter
	ld	hl,$+6		;10
	jp	p_enter		;10
.endm

p_enter::
	pushr_sp		;61	push current VM PC on rstack
	ld	sp,hl		;4	load VM PC
	ret			;10	next
#endif ; !VM_BYTE_OFFSETS

#if VM_INCLUDE_LONG || VM_BYTE_OFFSETS
;.macro p_enter_hlde
;	ld	ix,$+7		;14
;	jp	p_enter_ix	;10
;.endm

;p_enter_ix::
;	pushr_sp		;61	push current VM PC on rstack
;	ld	sp,ix		;8	load VM PC
;	ret			;10	next

.macro p_enter_hlde
	jp	p_enter_hlde	;10
.endm
	
p_enter_hlde::
	exx			;4	
	;push current VM PC on rstack:
	ld	hl,0		;10	
	add	hl,sp		;11	hl = VM PC
	ld	a,l		;4
	ld	(bc),a		;7
	inc	c		;4
	ld	a,h		;4
	ld	(bc++),a	;13
	
	;pop proc addr into VM PC, skip 'jp p_enter' and resume:
	dec	sp		;6
	dec	sp		;6 	sp -> proc address again
	pop	hl		;10 	hl -> proc
	ld	sp,hl		;4	sp -> proc
	inc	sp		;6	
	pop	af		;10	sp -> behind 'jp p_enter'
	exx			;4
	ret			;10	jump to next VM opcode
#endif ; VM_INCLUDE_LONG || VM_BYTE_OFFSETS


; ________________________________
; Enter z80 coded procedure which needs stack (and bc)
; the VM PC (register sp) is saved in sp_safe.
;
; there are variants which preserve hlde or de only,
; and variants as inline macro or helpers for jumping to.
; transparent helpers put a return address on stack (which may interfere with argument popping).
;
; • ret addr of finalizer on stack is no problem:
;   proc can simply return with 'ret'
;
; foo:	jp  enter_wstack_auto		; *must* be first instr in VM proc!
;	...
;	ret
;
; • ret addr of finalizer would interfere access to more arguments or slightly faster:
;
; foo:	jp  enter_wstack		; *must* be first instr in VM proc!
;	...
;	jp  return_wstack
;
; • faster and/or mixed requirement to preserve hlde and de only:
;
; foo:	ld_sp_vsp_via_hl
;	...
;	jp  return_wstack_hlde
;
; • function needs stack only in some cases:
;
; foo:	...
;	ld	hl, $+6
;	jp	jphl_wstack_auto
;	...
;	ret
;
; foo:	...
;	ld_sp_vsp_via_hl
;	...
;	jp	return_wstack
;
; Speed vs. Size:
;
;  macro + jp	de	10 bytes, 101cc	
;  enter + jp	de	 6 bytes, 167cc	
;  enter auto	de	 4 bytes, 178cc	
;  jphl_wstack…	de	 7 bytes, 148cc
;
;  macro + jp	hlde	13 bytes, 124cc	
;  enter + jp	hlde	 6 bytes, 214cc	
;  enter auto	hlde	 4 bytes, 225cc	
;  jpix_wstack…	hlde	 8 bytes, 183cc
;
; enter_wstack_only	 ( ade -- ade )	 	3 bytes: 88cc	*must* be first instr in VM proc!
; enter_wstack_only_hlde ( hlde -- hlde )	3 bytes: 124cc	*must* be first instr in VM proc!
; enter_wstack	  	 ( ade -- ade )	 	3 bytes: 178cc  *must* be first instr in VM proc!
; enter_wstack_hlde  	 ( hlde -- hlde )	3 bytes: 225cc	*must* be first instr in VM proc!
;
; .macro ld_sp_vsp_via_hl  ( ade -- ade )	 7 bytes: 32cc
; .macro ld_sp_vsp_via_ix  ( hlde -- hlde )	10 bytes: 44cc
; .macro ld_sp_vsp_via_ram ( hldeix -- hldeix )	12 bytes: 60cc
;
; .macro ld_vsp_sp_via_hl  ( ade -- ade )	10 bytes: 49cc
; .macro ld_vsp_sp_via_ram ( ahlde -- ahlde )	12 bytes: 60cc
;
; return_wstack      	( ade -- ade )		59cc
; return_wstack_hlde 	( hldeix -- hldeix )	70cc
;
; jphl_wstack	 	( de -- de )		128cc
; jpix_wstack_hlde   	( hlde -- hlde )	159cc

#if VM_FAST_CORE
;.macro enter_wstack_only
;	ld_sp_vsp_via_hl
;.endm
;
;.macro enter_wstack_only_hlde
;	ld_sp_vsp_via_ix
;.endm

.macro enter_wstack
	ld_sp_vsp_via_hl
	ld	hl, return_wstack
	push	hl
.endm

.macro enter_wstack_hlde
	ld_sp_vsp_via_ix
	ld	ix, return_wstack
	push	ix
.endm
#else ; !VM_FAST_CORE
;.macro enter_wstack_only
;	jp	enter_wstack_only
;.endm
;
;.macro enter_wstack_only_hlde
;	jp	enter_wstack_only_hlde
;.endm

.macro enter_wstack
	jp	enter_wstack
.endm

.macro enter_wstack_hlde
	jp	enter_wstack_hlde
.endm
#endif ; !VM_FAST_CORE


;enter_wstack_only::
;	dec	sp		;6
;	dec	sp		;6 	sp -> proc address again
;	pop	hl		;10 	hl -> proc
;	inc	hl		;6
;	inc	hl		;6
;	inc	hl		;6
;	ld_sp_vsp_via_ix	;44
;	jp	hl		;4	= 88cc	resume after 'jp enter_wstack'

enter_wstack::
	dec	sp		;6
	dec	sp		;6 	sp -> proc address again
	pop	hl		;10 	hl -> proc
	inc	hl		;6
	inc	hl		;6
	inc	hl		;6
jphl_wstack::
	ld_sp_vsp_via_ix	;44	
	rst	jp_hl		;25

ld_vsp_sp_via_hl_next::		; 	pres. de
return_wstack::			;
	ld_vsp_sp_via_hl	;49
	ret			;10	= 168cc

#if VM_INCLUDE_LONG
;enter_wstack_only_hlde::
;	dec	sp		;6
;	dec	sp		;6 	sp -> proc address again
;	pop	ix		;14 	ix -> proc
;	inc	ix		;10
;	inc	ix		;10
;	inc	ix		;10	ix -> behind 'jp enter_wstack'
;	ld_sp_vsp_via_ram	;60
;	jp	ix		;8	= 124cc

enter_wstack_hlde::
	dec	sp		;6
	dec	sp		;6 	sp -> proc address again
	pop	ix		;14 	ix -> proc
	inc	ix		;10
	inc	ix		;10
	inc	ix		;10

jpix_wstack_hlde::
	ld_sp_vsp_via_ram	;60
	rst	jp_ix		;29

ld_vsp_sp_via_ram_next::	; 	pres hlde
return_wstack_hlde::		;
	ld_vsp_sp_via_ram	;60
	ret			;10	= 215cc
#endif ; VM_INCLUDE_LONG


; _______________________________
; call proc ptr
;
; Convention:
; Procedures and VM words are called in the same way: VM words can be called like procs!
; In:  last argument is void or in de or in hlde.
; Out: return value  is void or in de or in hlde.
;
; CALLPROCPTR        (       de:addr -- hlde )
; CALLPROCPTR16      ( … n16 de:addr -- hlde )
; CALLPROCPTR16_hlde ( … hl:n16 de:addr -- hlde )
; CALLPROCPTR32      ( … n32 de:addr -- hlde )

CALLPROCPTR16::
	pop_hl
	;jr 	CALLPROCPTR16_hlde

CALLPROCPTR16_hlde::
CALLPROCPTR::
	ex	hl,de	; hl -> proc
	jp	hl	; jump there: proc is entered in mcode!

#if VM_INCLUDE_LONG
CALLPROCPTR32::
	ld	ix,de
	pop_de
	pop_hl
	jp	ix
#endif


; _______________________________
; return from procedure:
;
; RETURN       ( hlde -- hlde )
; DROPx_RETURN ( hlde -- hlde )

DROPx_RETURN::		; ~92cc
	dec	sp
	pop	af	; A = 2*N	
	jr	dropareturn
	
DROP8_RETURN::		; ~83cc
	ld	a,16		
	jr	dropareturn

DROP7_RETURN::		; ~83cc
	ld	a,14
	jr	dropareturn

DROP6_RETURN::		; ~83cc
	ld	a,12
	jr	dropareturn

DROP5_RETURN::		; ~83cc
	ld	a,10
	jr	dropareturn

DROP4_RETURN::		; ~83cc
	ld	a,8
	jr	dropareturn

DROP3_RETURN::		; ~83cc
	ld	a,6	
	;jr	dropareturn	

dropareturn:		; ~64cc
	add	a,c	;4
	ld	c,a	;4
	popr_sp		;44	pres. cy
	ret	nc	;5|11
	inc	b	;4
	ret		;10

DROP2_RETURN::		; 74cc
	inc	c
	inc	bc

DROP1_RETURN::		; 64cc
	inc	c
	inc	bc

DROP0_RETURN::
RETURN::		; 54cc
	popr_sp		;44
	ret		;10


; ________________________________
; JP opcodes
; 
; JP opcodes test the interrupt state

; JP      ( de $addr.w -- de )		; jp to label
; JP_hlde ( hlde $addr.w -- hlde )	; jp to label
; JP0     ( de:n $addr.w -- void )	; jp to label if n==0
; JP1     ( de:n $addr.w -- void )	; jp to label if n!=0
;
; AND0JP  ( n $addr.w -- void )	 	; if n==0 keep n and jp to label (TOP=n)
;				 	  else    drop n & don't jp      (TOP=VOID)
; OR1JP   ( n $addr.w -- void )   	; if n!=0 keep n and jp to label (TOP=n)
;				 	  else    drop n & don't jp      (TOP=VOID)

JP::	pop	hl		;10	HL := $addr.w
	ld	sp,hl		;4	VM PC := HL
	in	a, k1_rd_irpt	;8
	inc	a		;4
	ret	z		;5|11	= 37cc	
irpt:	ei
	ld	sp, irpt_stack
	di
	ld	sp,hl
	ret

POLL_IRPT::
	in	a, k1_rd_irpt	;8
	inc	a		;4
	ret	z		;5|11	= 23cc
	ld	hl,0	
	add	hl,sp	
	jr	irpt

JP_hlde::
	pop	ix		;14	HL := $addr.w
	ld	sp,ix		;8	VM PC := HL
	in	a, k1_rd_irpt	;8
	inc	a		;4
	ret	z		;5|11	= 45cc	
	ei
	ld	sp, irpt_stack
	di
	ld	sp,ix
	ret

AND0JP::
JP0::	ld	a,d		;4
	or	e		;4
	pop	hl		;10
	ret	nz		;5|11	= 29cc	don't jump
	ld	sp,hl		;4
	in	a, k1_rd_irpt	;8
	inc	a		;4
	ret	z		;5|11	= 50cc	jump
	jr	irpt

OR1JP::
JP1::	ld	a,d
	or	e
	pop	hl
	ret	z		; don't jump
	ld	sp,hl
	in	a, k1_rd_irpt	
	inc	a		
	ret	z		; jump
	jr	irpt


; ________________________________
; if n<L[].count jump to label L[n] else jump to default label
; jump table can have from 1 to 256 entries for case 0 to 255
;
; SWITCH ( de:n $tablesize.b $defaultlabel.w $table.w[] -- void )
;
; invocation:
;
;	dw	SWITCH
;	db	tablesize-1
;	dw	default_label
;	dw	table[tablesize]; list of labels

SWITCH::
	dec	sp
	pop	af		; a = tablesize-1, sp -> default_label

	rlc	d		; test de.hi
	jr	nz,1$		; de.hi != 0 -> oorange

	cp	a,e		; sz-1 - idx
	jr	c,1$		; sz-1<idx -> sz≤idx -> idx≥sz -> beyond table
	
	; de inside table
	ex	hl,de		; hl=idx
	inc	hl		; +1 to skip over default entry
	add	hl,hl		; hl=idx*2 for word-size entries
	add	hl,sp		; add table base
	ld	sp,hl		; sp -> label_n

1$:	pop	hl		; hl = default_label or label_n
	ld	sp,hl
	ret


; -----------------------------------------
; 	PUSH IMMEDIATE VALUE
; -----------------------------------------

; ________________________________
; IVAL    ( void $.w -- de:n16 )	; push word value 
; IVALm1  ( void -- de:n16 )	; push immediate value -1
; IVAL0   ( void -- de:n16 )	; push immediate value 0
; IVAL1   ( void -- de:n16 )	; push immediate value 1
; IVAL2   ( void -- de:n16 )	; push immediate value 2
; IVAL3   ( void -- de:n16 )	; push immediate value 3
; IVAL4   ( void -- de:n16 )	; push immediate value 4


PUSH32_IVAL::	push_hl		; ( hlde -- n )
PUSHDE_IVAL::	push_de		; ( de -- n )
		pop	de
		ret

EXHLDE_IVAL::	ex	hl,de	; ( de -- hl n )
IVAL::		pop  	de	; ( void -- n )
		ret

;PUSHDE_IVALm1:: push_de
EXHLDE_IVALm1::	ex	hl,de	; AUX = TOP
IVALm1::	ld	de,-1	; TOP = -1
		ret

;PUSHDE_IVAL0::	push_de
EXHLDE_IVAL0::	ex	hl,de	; AUX = TOP
IVAL0::		ld	de,0	; TOP = 0
		ret

;PUSHDE_IVAL1::	push_de
EXHLDE_IVAL1::	ex	hl,de
IVAL1::		ld	de,1
		ret

;PUSHDE_IVAL2::	push_de
EXHLDE_IVAL2::	ex	hl,de
IVAL2::		ld	de,2
		ret

;PUSH_IVAL3::	push_de
EXHLDE_IVAL3::	ex	hl,de
IVAL3::		ld	de,3
		ret

;PUSH_IVAL4::	push_de
EXHLDE_IVAL4::	ex	hl,de
IVAL4::		ld	de,4
		ret

#if VM_BYTE_OFFSETS ; small VM program wanted
;PUSHDE_IVALu8:: push_de
EXHLDE_IVALu8::	ex	hl,de
IVALu8::	pop	de
		dec	sp
		ld	d,0
		ret

;PUSHDE_IVALs8:: push_de
EXHLDE_IVALs8::	ex	hl,de
IVALs8::	pop	de
		dec	sp
		ld	d,0xff
		ret
#endif ; VM_BYTE_OFFSETS


; ________________________________
; IVAL32  ( void $.l -- hlde:n32 )	; push long value
; IVAL32u ( void $.w -- hlde:n32 )	; push positive long value (hi word = 0x0000)
; IVAL32s ( void $.w -- hlde:n32 )	; push negative long value (hi word = 0xffff)

PUSH32_IVAL32::	push_hl		; ( hlde -- hlde:n32 )
PUSHDE_IVAL32::	push_de		; ( de   -- hlde:n32 )
IVAL32::	pop  	de	; ( void -- hlde:n32 )
		pop  	hl	; hl=n.hi, de=n.lo
		ret

;PUSH32_IVAL32s::push_hl 	; ( hlde -- hlde:n32 )
;PUSHDE_IVAL32s::push_de 	; ( de   -- hlde:n32 )
IVAL32s::	ld	hl,-1	; ( void -- hlde:n32 )
		pop	de	; hl=n.hi, de=n.lo
		ret
	
PUSH32_IVAL32u::push_hl		; ( hlde -- hlde:n32 )
PUSHDE_IVAL32u::push_de		; ( de   -- hlde:n32 )
IVAL32u::	ld	hl,0	; ( void -- hlde:n32 )
		pop	de	; hl=n.hi, de=n.lo
		ret


; -----------------------------------------
; 	MOVE TO / FROM RSTACK
; -----------------------------------------

; TOR     ( de -- void )	; move word from vstack to rstack
; DUP2R   ( de -- de )		; copy top value from vstack to rstack
; FROMR   ( void -- de )	; move word from rstack to vstack
; PEEKR   ( void -- de )	; copy top value from rstack to vstack
;
; TOR32   ( hlde -- void )	; move int32 or two int16 to rstack
; DUP2R32 ( hlde -- hlde )	; copy top long value to rstack
; FROMR32 ( void -- hlde )	; move int32 from rstack to vstack


DUP2R::				; ( de:n -- de:n )
TOR::		pushr_de	; ( de:n -- void )
		ret

PUSHDE_FROMR::	push_de		; ( de:top -- de:n )
		ex	hl,de	; else hl will become void
EXHLDE_FROMR::	ex	hl,de	; ( de:top -- hl:aux de:n )
FROMR::		popr_de		; ( de:void -- de:n )
		ret

PUSHDE_PEEKR::	push_de		; ( de:top -- de:n )
		ex	hl,de	; else hl will become void
EXHLDE_PEEKR::	ex	hl,de	; ( de:top -- hl:aux de:n )
PEEKR::		exx		; ( void -- de:n )
		ld	a,(--bc)
		exx
		ld	d,a
		exx
		dec	c
		ld	a,(bc)
		inc	c
		inc	bc
		exx
		ld	e,a
		ret

DUP2R32::			; ( hlde -- hlde )
TOR32::		pushr_de	; ( hlde -- void )
		pushr_hl
		ret

FROMR32::	popr_hl		; ( void -- hlde )
		popr_de
		ret

;PEEKR32::	popr_hl		; ( void -- hlde )
;		popr_de
;		exx		
;		inc	bc
;		inc	bc
;		inc	bc
;		inc	bc
;		exx
;		ret



; -----------------------------------------
; 	STACK BASICS
; -----------------------------------------

; ________________________________
; DUP   ( de:n16 -- n16 de:n16 )	; LGET 0;   duplicate TOP value
; DUP32 ( hlde:n32 -- n32 hlde:n32 )	; LGET32 0; duplicate AUX+TOP long value
; SWAP  ( n1 de:n2 -- n2 de:n1 ) 	; swap two words on stack


;DUP::		equ	PUSHDE		; ( de:n -- n de:n )
;DUP32::	equ	PUSH32		; ( hlde:n32 -- hlde:n32 )

;SWAP::		ld	hl,bc
;		ld	a,(hl)
;		ld	(hl),e
;		ld	e,a
;		inc	hl
;		ld	a,(hl)
;		ld	(hl),d
;		ld	d,a
;		ret

; ________________________________
; DROP	     ( hlde:a -- void )		; drop top value in DE or HLDE
; DROP2      ( b hlde:a -- void )	; drop 2 words
; DROP3      ( c b hlde:a -- void )	; drop 3 words
; DROP4      ( d c b hlde:a -- void )	; drop 4 words
; DROPx      ( … $n.b -- void )		; drop n-1 words
;
; NIP and DROP are identical, with the exception that after DROP HLDE is considered void.
; To avoid confusion, only DROP is used and DROP drops the given number of words from the STACK 
; ignoring (and preserving) DE and HL. NOTE: this is different to the original DROP behaviour.

DROPx::	dec	sp		;6	
	pop	af		;10	a = N*2
	jr	dropa		;12+20	~48cc

DROP8::	inc	c		;	~37cc
	inc	bc
DROP7::	ld	a,14		;7	~27cc
dropa:	add	a,c		;4	~20cc
	ld	c,a		;4
	ret	nc		;5|11
	inc	b		;4
	ret			;10

DROP6::	inc	c		;	~37cc
	inc	bc
DROP5::	ld	a,10		;7	~27cc
	add	a,c		;4
	ld	c,a		;4
	ret	nc		;5|11
	inc	b		;4
	ret			;10

DROP4::	inc	c		;4	~37cc
	inc	bc		;6
DROP3::	ld	a,6		;7	~27cc
	add	a,c		;4
	ld	c,a		;4
	ret	nc		;5|11
	inc	b		;4
	ret			;10

DROP2::	inc	c		;	30cc
       	inc	bc

DROP1::
DROP::	inc	c		;4	20cc
       	inc	bc		;6
DROP0::	ret			;10


#endlocal

#include "core-k1bus.s"		; K1-Bus related
#include "core-vars.s"		; read/write local/global variables, struct or array members
#include "core-int8.s"		; words working on int8 / uint8 values
#include "core-int16.s"		; words working with int16 / uint16 values
#include "core-int32.s" 	; words working with int32 / uint32 values
;#include "core-float32.s"
#include "core-memory.s"	; heap memory
#include "core-strings.s"	; arrays and strings
#include "core-misc.s"		; some functions


